#include <config.h>

#include <math.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <libgnomeui/gnome-about.h>
#include <libgnomeui/gnome-window-icon.h>
#include <panel-applet.h>

/* sonypi stuff */
#include <linux/types.h>
// Define our ioctl's (stolen from spicctrl)
#define SONYPI_IOCGBRT _IOR('v', 0, __u8)
#define SONYPI_IOCSBRT _IOW('v', 0, __u8)

typedef struct _VaioData {
        
    gint brightness_before_popup;
    gint brightness;

    PanelAppletOrient orientation;    
        
    GtkAdjustment *adj;
    GtkWidget *applet;
    GtkWidget *frame;
    GtkWidget *image;
    
    /* popup window and scale */
    GtkWidget *popup;
    GtkWidget *scale;
    
    /* various icons */
    GdkPixbuf *zero;
    GdkPixbuf *min;
    GdkPixbuf *mid;
    GdkPixbuf *max;

    GtkTooltips *tooltips;
    
    /* Sony Programmable Interface */
    gint sony_fd;
    
} VaioData;

#define VAIOLCD_MAX 255
#define VAIOLCD_MID 127
#define VAIOLCD_MIN 63
#define VAIOLCD_ICON_MAX "/gnome-vaiolcd-max.png"
#define VAIOLCD_ICON_MID "/gnome-vaiolcd-mid.png"
#define VAIOLCD_ICON_MIN "/gnome-vaiolcd-min.png"
#define VAIO_SONYPI "/dev/sonypi"

/* constants */
static const gchar *access_name = N_("Vaio LCD Brightness Control");

/* function prototypes */
static gint vaiolcd_get_brightness(VaioData *data);
static void vaiolcd_set_brightness(VaioData *data, gint value);

/* static void vaiolcd_init_images (); */
static void vaiolcd_load_images (VaioData *data);
static void vaiolcd_update_slider (VaioData *data);
static void vaiolcd_update_image (VaioData *data);
static void vaiolcd_popup_show (VaioData *data);
static void vaiolcd_popup_hide (VaioData *data, gboolean revert);

/* function prototypes - ui callbacks */
static gboolean applet_button_release_event_cb (GtkWidget *widget, GdkEventButton *event, VaioData *data);
static gboolean applet_key_press_event_cb (GtkWidget *widget, GdkEventKey *event, VaioData *data);
static void applet_style_event_cb (GtkWidget *widget, GdkEventButton *event, VaioData *data);
static gboolean applet_scroll_event_cb (GtkWidget *widget, GdkEventScroll *event, VaioData *data);
static void applet_change_orient_cb (GtkWidget *widget, PanelAppletOrient o, VaioData *data);
static void applet_change_size_cb (GtkWidget *widget, gint size, VaioData *data);
static void applet_change_background_cb (PanelApplet *applet,
                                             PanelAppletBackgroundType  type,
                                             GdkColor *color,
                                             const gchar *pixmap,
                                             VaioData *data);

static gboolean scale_button_release_event_cb (GtkWidget *widget, GdkEventButton *event, VaioData *data);
static gboolean scale_key_press_event_cb (GtkWidget *widget, GdkEventKey *event, VaioData *data);


static void vaiolcd_value_changed_cb(GtkAdjustment *adj, VaioData *data);
static void vaiolcd_destroy_cb(GtkWidget *widget, VaioData *data);

/* function prototypes - menu callbacks */
static void vaiolcd_properties_cb (BonoboUIComponent *uic,
                              VaioData         *data,
                              const gchar       *verbname);
static void vaiolcd_about_cb (BonoboUIComponent *uic,
                              VaioData         *data,
                              const gchar       *verbname);



/* menu callbacks */
const BonoboUIVerb vaiolcd_applet_menu_verbs [] = {
/*    BONOBO_UI_UNSAFE_VERB ("Properties", vaiolcd_properties_cb), */
    BONOBO_UI_UNSAFE_VERB ("About", vaiolcd_about_cb),
    BONOBO_UI_VERB_END
};


static gint vaiolcd_open_device (VaioData *data) {
    if (!data->sony_fd) {
        data->sony_fd = open(VAIO_SONYPI,O_RDWR);
        if (!data->sony_fd) {
            data->sony_fd = 0;
            return -1;
        }
    }
    return 0;
}

static void vaiolcd_close_device(VaioData *data) {
    if (data->sony_fd)
        close(data->sony_fd);
}

static gint vaiolcd_get_brightness(VaioData *data) {

    __u8 value = 0;
    
    if (!data->sony_fd) {
        if (vaiolcd_open_device(data) == -1)
            return VAIOLCD_MAX;
    }
    
    if (ioctl(data->sony_fd, SONYPI_IOCGBRT, &value) < 0) {
        return VAIOLCD_MAX;
    }

    return value;
}

static void  vaiolcd_set_brightness(VaioData *data, gint value) {
    __u8 v = (__u8)value;
    if (!data->sony_fd) {
        if (vaiolcd_open_device(data) == -1)
            return;
    }
    
    if (ioctl(data->sony_fd, SONYPI_IOCSBRT, &v) < 0) {
        return;
    }
}

static void vaiolcd_load_images (VaioData *data) {
    /* FIXME - load all icons using gdk_pixbuf_from_image() rather
     * than going through making an icon factory
     * FIXME - do all the other icons
     */
	GError *error = NULL;
	
    if (data->max)
        g_object_unref (data->max);
	data->max = gdk_pixbuf_new_from_file(VAIOLCD_ICONDIR VAIOLCD_ICON_MAX, &error);
	
	if (data->mid)
		g_object_unref (data->mid);
	data->mid = gdk_pixbuf_new_from_file(VAIOLCD_ICONDIR VAIOLCD_ICON_MID, &error);

	if (data->min)
		g_object_unref (data->min);
	data->min = gdk_pixbuf_new_from_file(VAIOLCD_ICONDIR VAIOLCD_ICON_MIN, &error);
	
}

static void vaiolcd_update_slider(VaioData *data) {
    gint brightness = data->brightness;
    brightness = -brightness;
    gtk_adjustment_set_value(data->adj, brightness);
}

static void vaiolcd_update_image(VaioData *data) {
    gint brightness, size;
    GdkPixbuf *pixbuf, *copy, *scaled = NULL;
    
    brightness = data->brightness;
    
    size = panel_applet_get_size (PANEL_APPLET (data->applet)) - 6;
    
	if (brightness < VAIOLCD_MIN)
		pixbuf = gdk_pixbuf_copy(data->min);
	else if (brightness < VAIOLCD_MID)
		pixbuf = gdk_pixbuf_copy(data->mid);
	else
		pixbuf = gdk_pixbuf_copy(data->max);
    
    scaled = gdk_pixbuf_scale_simple (pixbuf, size, size, GDK_INTERP_BILINEAR);
    gtk_image_set_from_pixbuf (GTK_IMAGE (data->image), scaled);
    g_object_unref (scaled);
    g_object_unref (pixbuf);	
	/*
    gtk_widget_set_size_request (GTK_WIDGET (data->frame),
                                 MAX(11, size),
                                 MAX(11, size));
	 */
}

static void vaiolcd_popup_show (VaioData *data) {
    
    GtkRequisition req;
    GtkWidget *frame;
    GtkWidget *inner_frame;
    gint x,y;
    gint width, height;
    GdkGrabStatus pointer, keyboard;
    
    data->popup = gtk_window_new (GTK_WINDOW_POPUP);
    gtk_window_set_screen (GTK_WINDOW (data->popup),
                           gtk_widget_get_screen (data->applet));
    
    data->brightness_before_popup = vaiolcd_get_brightness (data);

    frame = gtk_frame_new (NULL);
    gtk_container_set_border_width (GTK_CONTAINER (frame), 0);
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
    gtk_widget_show (frame);
    
    inner_frame = gtk_frame_new (NULL);
    gtk_container_set_border_width (GTK_CONTAINER (inner_frame), 0);
    gtk_frame_set_shadow_type (GTK_FRAME (inner_frame), GTK_SHADOW_NONE);
    
    if (data->orientation == PANEL_APPLET_ORIENT_UP ||
        data->orientation == PANEL_APPLET_ORIENT_DOWN ) {
        data->scale = gtk_vscale_new (data->adj);
        gtk_widget_set_size_request (data->scale, -1, 100);
    }
    else {
        data->scale = gtk_hscale_new (data->adj);
        gtk_widget_set_size_request (data->scale, 100, -1);
        gtk_range_set_inverted (GTK_RANGE (data->scale), TRUE);
    }
    
    /* FIXME Gail / Accessibility */
    
    g_signal_connect (data->scale,
                      "button-release-event",
                      (GCallback) scale_button_release_event_cb,
                      data);
    g_signal_connect (data->scale,
                      "key-press-event",
                      (GCallback) scale_key_press_event_cb,
                      data);
    
    gtk_widget_show (data->scale);
    gtk_scale_set_draw_value (GTK_SCALE (data->scale), FALSE);
    gtk_range_set_update_policy (GTK_RANGE (data->scale), GTK_UPDATE_CONTINUOUS);
    
    gtk_container_add (GTK_CONTAINER (data->popup), frame);
    gtk_container_add (GTK_CONTAINER (frame), inner_frame);
    gtk_container_add (GTK_CONTAINER (inner_frame), data->scale);
    
    gtk_widget_show_all (inner_frame);
    
    
    /* position the popup next to the button */
    gtk_widget_size_request (data->popup, &req);
    gdk_window_get_origin (data->image->window, &x, &y);
    gdk_drawable_get_size (data->image->window, &width, &height);
    
    switch (data->orientation) {
    case PANEL_APPLET_ORIENT_DOWN:
        x += (width - req.width) / 2;
        y += height;
        break;
        
    case PANEL_APPLET_ORIENT_UP:
        x += (width - req.width) / 2;
        y -= req.height;
        break;
        
    case PANEL_APPLET_ORIENT_LEFT:
        x -= req.width;
        y += (height - req.height) / 2;
        break;
        
    case PANEL_APPLET_ORIENT_RIGHT:
        x += width;
        y += (height - req.height) / 2;
        break;
        
    default:
        break;
    }
    
    if (x < 0)
        x = 0;
    if (y < 0)
        y = 0;
    
    gtk_window_move (GTK_WINDOW (data->popup), x, y);
    gtk_widget_show (data->popup);
    
    /* grab focus and pointer */
    gtk_widget_grab_focus (data->scale);
    gtk_grab_add (data->scale);
    
    pointer = gdk_pointer_grab (data->scale->window,
                                TRUE,
                                (GDK_BUTTON_PRESS_MASK 
                                 | GDK_BUTTON_RELEASE_MASK
                                 | GDK_BUTTON_MOTION_MASK),
                                NULL,
                                NULL,
                                GDK_CURRENT_TIME);
    
    keyboard = gdk_keyboard_grab (data->scale->window,
                                  TRUE,
                                  GDK_CURRENT_TIME);
    
    if (pointer != GDK_GRAB_SUCCESS || keyboard != GDK_GRAB_SUCCESS) {
        /* failed to grab keyboard/mouse - refuse to show slider*/
        gtk_grab_remove (data->scale);
        gtk_widget_destroy (data->popup);
        data->popup = NULL;
        data->scale = NULL;
        
        if (pointer == GDK_GRAB_SUCCESS)
            gdk_keyboard_ungrab (GDK_CURRENT_TIME);
        if (keyboard == GDK_GRAB_SUCCESS)
            gdk_pointer_ungrab (GDK_CURRENT_TIME);
        
        g_warning ("vaiolcd_applet: Could not grab X server.");
        return;
    }
    
    gtk_frame_set_shadow_type (GTK_FRAME (data->frame), GTK_SHADOW_IN);
    
}

static void vaiolcd_popup_hide (VaioData *data, gboolean revert) {
    
    if (data->popup) {
        gtk_grab_remove (data->scale);
        gdk_pointer_ungrab (GDK_CURRENT_TIME);
        gdk_keyboard_ungrab (GDK_CURRENT_TIME);
        
        /* Ref the adjustment or it will get destroyed when scale
         * goes away
         */
        g_object_ref (G_OBJECT (data->adj));
        
        gtk_widget_destroy (GTK_WIDGET (data->popup));
        
        data->popup = NULL;
        data->scale = NULL;
        
        if (revert) {
            vaiolcd_set_brightness (data, data->brightness_before_popup);
        }
        
        gtk_frame_set_shadow_type (GTK_FRAME (data->frame), GTK_SHADOW_NONE);
    }
}
        
static gboolean applet_button_release_event_cb (GtkWidget *widget, 
                                            GdkEventButton *event, 
                                            VaioData *data) {
    
    if (event->button == 1) {
        if (data->popup == NULL) {
            vaiolcd_popup_show (data);
            return TRUE;
        }
    }
    
    return TRUE;
}

static gboolean applet_key_press_event_cb (GtkWidget *widget, 
                                       GdkEventKey *event, 
                                       VaioData *data) {
    
    switch (event->keyval) {
    case GDK_Escape:
        /* revert */
        vaiolcd_popup_hide (data, TRUE);
        return TRUE;
    case GDK_KP_Enter:
    case GDK_ISO_Enter:
    case GDK_3270_Enter:
    case GDK_Return:
    case GDK_KP_Space:
        /* apply */
        if (data->popup != NULL)
            vaiolcd_popup_hide (data, FALSE);
        else {
			vaiolcd_update_slider(data);
            vaiolcd_popup_show (data);
		}
        return TRUE;
    default:
        break;
    }
    
    return FALSE;
}

static void applet_style_event_cb (GtkWidget *widget, 
                                   GdkEventButton *event, 
                                   VaioData *data) {
    vaiolcd_load_images (data);
    vaiolcd_update_image (data);
}

static gboolean applet_scroll_event_cb (GtkWidget *widget,
                                    GdkEventScroll *event,
                                    VaioData *data) {
    
    gint direction;
    if (event->type != GDK_SCROLL ) {
        return FALSE;
    }
    
    direction = event->direction;
    
    switch (direction) {
    case GDK_SCROLL_UP:
        data->brightness += 5;
        break;
    case GDK_SCROLL_DOWN:
        data->brightness -= 5;
        break;
    default:
        break;
    }
    
    if (data->brightness > VAIOLCD_MAX) data->brightness = VAIOLCD_MAX;
    if (data->brightness < 0 ) data->brightness = 0;
    
    vaiolcd_update_slider (data);
    return TRUE;
}

static void applet_change_orient_cb (GtkWidget *widget, 
                                     PanelAppletOrient o, 
                                     VaioData *data) {
    vaiolcd_popup_hide (data, FALSE);
    data->orientation = o;
}

static void applet_change_size_cb (GtkWidget *widget, 
                                   gint size, 
                                   VaioData *data) {
    vaiolcd_popup_hide (data, FALSE);
    vaiolcd_update_image (data);
}

static void applet_change_background_cb (PanelApplet *applet,
                                         PanelAppletBackgroundType  type,
                                         GdkColor *color,
                                         const gchar *pixmap,
                                         VaioData *data) {
 
    if (type == PANEL_NO_BACKGROUND) {
        GtkRcStyle *rc_style = gtk_rc_style_new ();
        gtk_widget_modify_style (data->applet, rc_style);
        gtk_rc_style_unref (rc_style);
    }
    else if (type == PANEL_COLOR_BACKGROUND) {
        gtk_widget_modify_bg (data->applet, GTK_STATE_NORMAL, color);
    }
    else {
        /* FIXME: pixmap */
    }
}

static gboolean scale_button_release_event_cb (GtkWidget *widget, 
                                           GdkEventButton *event, 
                                           VaioData *data) {
    if (data->popup != NULL) {
        vaiolcd_popup_hide (data, FALSE);
        return TRUE;
    }
    
    return FALSE;
}

static gboolean scale_key_press_event_cb (GtkWidget *widget, 
                                      GdkEventKey *event,
                                      VaioData *data) {
    	switch (event->keyval) {
	case GDK_Escape:
		/* Revert. */
		vaiolcd_popup_hide (data, TRUE);
		return TRUE;

	case GDK_KP_Enter:
	case GDK_ISO_Enter:
	case GDK_3270_Enter:
	case GDK_Return:
	case GDK_space:
	case GDK_KP_Space:
		/* Apply. */
		vaiolcd_popup_hide (data, FALSE);
		return TRUE;

	default:
		break;
	}

	return FALSE;
}

static void vaiolcd_value_changed_cb(GtkAdjustment *adj, VaioData *data) {
    
    gint brightness;
    brightness = -1 * (adj->value);
    data->brightness = brightness;
    vaiolcd_update_image (data);

    vaiolcd_set_brightness (data, brightness);
}


static void vaiolcd_about_cb (BonoboUIComponent *uic,
                              VaioData         *data,
                              const gchar       *verbname) {
    
    static GtkWidget   *about     = NULL;
    static const gchar *authors[] = {
        "Alastair Tse <liquidx@liquidx.net>",
            NULL
    };

	const gchar *documenters[] = {
		NULL
	};

	const gchar *translator_credits = _("translator_credits");
	
    if (about) {
		gtk_window_set_screen (GTK_WINDOW (about),
                               gtk_widget_get_screen (data->applet));
        gtk_window_present (GTK_WINDOW (about));
        return;
    }
    
    /*
    file = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_PIXMAP, "mixer/gnome-mixer-applet.png", TRUE, NULL);
    pixbuf = gdk_pixbuf_new_from_file (file, NULL);
    g_free (file);
    */

    about = gnome_about_new (_("Vaio LCD Brightness Control Applet"), VERSION,
                             _("(C) 2003 Alastair Tse"),
                             _("Sets LCD Panel brightness for Sony 505 Laptops."),
                             authors,
                             documenters,
                             strcmp (translator_credits, "translator_credits") != 0 ? translator_credits : NULL,
                             NULL /*pixbuf*/);
    
	gtk_window_set_screen (GTK_WINDOW (about),
                           gtk_widget_get_screen (data->applet));
	gtk_window_set_wmclass (GTK_WINDOW(about), "vaio lcd brightness", "Vaio LCD Brightness");
	/* gnome_window_icon_set_from_file (GTK_WINDOW (about), GNOME_ICONDIR"/mixer/gnome-mixer-applet.png"); */
    g_signal_connect (G_OBJECT (about), "destroy",
                      G_CALLBACK (gtk_widget_destroyed),
	                  &about);
	
    gtk_widget_show (about);
}

static void vaiolcd_properties_cb (BonoboUIComponent *uic,
                              VaioData         *data,
                              const gchar       *verbname) {
    vaiolcd_about_cb(uic, data, verbname);
}


static void vaiolcd_destroy_cb (GtkWidget *widget, VaioData *data) {
    vaiolcd_close_device(data);
    g_free(data);
}

static void vaiolcd_applet_create (PanelApplet *applet) {
 
    VaioData *data;
    
    data = g_new0 (VaioData, 1);
    
    data->frame = gtk_frame_new (NULL);
    gtk_frame_set_shadow_type (GTK_FRAME(data->frame), GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER (applet), data->frame);
    
    data->applet = GTK_WIDGET(applet);
    
    data->image = gtk_image_new();
    gtk_container_add (GTK_CONTAINER (data->frame), data->image);
    
    gtk_tooltips_set_tip (data->tooltips,data->applet,
                          _(access_name),NULL);

    /* get initial brightness */
    data->brightness = vaiolcd_get_brightness (data);
	data->brightness_before_popup = data->brightness;
	vaiolcd_set_brightness(data, data->brightness);
	
    /* FIXME: gail / accessibility stuff */
    
    g_signal_connect (data->applet,
                      "button-release-event",
                      (GCallback) applet_button_release_event_cb,
                      data);
    
	g_signal_connect (data->applet,
                      "key-press-event",
                      (GCallback) applet_key_press_event_cb,
                      data);
	
	g_signal_connect (data->applet,
                      "scroll-event",
                      (GCallback) applet_scroll_event_cb,
                      data);

	g_signal_connect (data->applet,
                      "style-set",
                      (GCallback) applet_style_event_cb,
                      data);

    /* set slider */
    
    data->adj = GTK_ADJUSTMENT (gtk_adjustment_new(-50,
                                                   -VAIOLCD_MAX,
                                                   0.0,
                                                   VAIOLCD_MAX/20,
                                                   VAIOLCD_MAX/10,
                                                   0.0));
    g_signal_connect (data->adj,
                      "value-changed",
                      (GCallback) vaiolcd_value_changed_cb,
                      data);
    
    /* other callback handlers */
    g_signal_connect (applet,
                      "destroy",
                      G_CALLBACK (vaiolcd_destroy_cb),
                      data);
    g_signal_connect (applet,
                      "change_orient",
                      G_CALLBACK (applet_change_orient_cb),
                      data);
    
    g_signal_connect (applet,
                      "change_size",
                      G_CALLBACK (applet_change_size_cb),
                      data);
    
    g_signal_connect (applet,
                      "change_background",
                      G_CALLBACK (applet_change_background_cb),
                      data);
    
    panel_applet_setup_menu_from_file (applet,
                                       NULL, /*opt_datadir */
                                       "GNOME_VaioLCDApplet.xml",
                                       NULL,
                                       vaiolcd_applet_menu_verbs,
                                       data);
    
    applet_change_orient_cb (GTK_WIDGET(applet),
                             panel_applet_get_orient (applet),
                             data);
    
    applet_change_size_cb (GTK_WIDGET(applet),
                           panel_applet_get_size (applet),
                           data);
    
    vaiolcd_update_slider (data);
    vaiolcd_update_image (data);
    
    vaiolcd_open_device (data);
    
    gtk_widget_show_all (GTK_WIDGET (applet));
    
}

static gboolean vaiolcd_applet_factory (PanelApplet *applet,
                                       const gchar *iid,
                                       gpointer data) {
    vaiolcd_applet_create(applet);
    return TRUE;
}
    

PANEL_APPLET_BONOBO_FACTORY ( "OAFIID:GNOME_VaioLCDApplet_Factory",
                              PANEL_TYPE_APPLET,
                              "GNOME_VaioLCDApplet",
                              "0",
                              vaiolcd_applet_factory,
                              NULL);



